/*
 * Copyright (c) 2003-2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-net-cl.c,v 1.11 2013/09/18 03:15:46 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/test.h"
#include "sm/io.h"
#include "sm/ctype.h"
#include "sm/time.h"
#include "smi-net.h"
#include "sm/unixsock.h"

#include <stdio.h>

static char *nsocket = "./cltsrv";

#if 0
#include "timing.h"
#include "t-net-common.c"
#else /* 0 */
#define SM_DEFPORT	2525
static int Verbose;

void
usage(const char *prg)
{
	fprintf(stderr, "%s: write QUIT\r\n to localhost\n", prg);
	fprintf(stderr, "usage: %s [options]\n", prg);
#if MTA_NETINET6
	fprintf(stderr, "-6: use INET6 socket\n");
#endif
	fprintf(stderr, "-p port: port to use [default: %d]\n", SM_DEFPORT);
	fprintf(stderr, "-r n: read data first (n: timeout in seconds)\n");
	fprintf(stderr, "-s n: write slowly (n: total time to write in seconds)\n");
	fprintf(stderr, "-u: use UNIX socket instead of INET\n");
	fprintf(stderr, "-U socket: use Unix domain socket [default: %s]\n",
		nsocket);
	fprintf(stderr, "-V: increase verbosity\n");
	exit(0);
}
#endif /* 0 */

/*
**  CLIENT -- write "QUIT\r\n" to localhost:port
**	This is a hack to connect to a SMTP server to figure out whether
**	it is accepting connections (at least).
**
**	Parameters:
**		port -- port
**		family -- type of socket
**		readfirst -- read first some data
**		ctmo -- connect timeout
**		slow -- write slowly
**
**	Returns:
**		none
*/

#define toseconds(x, y)	(x.tv_sec - y.tv_sec)

int
client(int port, int family, int readfirst, int ctmo, int slow)
{
	int fd;
	sm_ret_T ret;
	char buf[] = "QUIT\r\n";
	struct timeval t1, t2;

	if (Verbose > 1)
		fprintf(stderr, "clt: connect\n");
	if (ctmo > 0) {
		if (gettimeofday(&t1, NULL) < 0)
			perror("gettimeofday");
	}
	while (1) {
		int delay;

		if (AF_INET == family)
			(void) net_client_connect("127.0.0.1", port, &fd);
#if MTA_NETINET6
		else if (AF_INET6 == family)
			(void) net_client_connect("::1", port, &fd);
#endif
		else if (AF_UNIX == family)
			(void) unix_client_connect(nsocket, &fd);
		if (ctmo <= 0)
			break;
		if (gettimeofday(&t2, NULL) < 0)
			perror("gettimeofday");
		delay = (int) toseconds(t2, t1);
		if (fd >= 0) {
			printf("connect time: %d seconds\n", delay);
			break;
		}
		if (delay >= ctmo) {
			printf("connection timed out, delay=%d\n", delay);
			break;
		}
		if (Verbose > 1)
			fprintf(stderr, "connect delay=%d, still trying\n", delay);
		sleep(1);
	}
	if (Verbose > 1)
		fprintf(stderr, "clt: connected=%d\n", fd);
	SM_TEST(fd >= 0);
	if (fd < 0)
		return 1;

	if (readfirst > 0) {
		int r, nfds;
		ssize_t rd;
		fd_set readfds;
		struct timeval timeout;
		char rdbuf[256];

		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		nfds = fd + 1;
		timeout.tv_usec = 0;
		timeout.tv_sec = readfirst;
		r = sm_fd_nonblock(fd, true);
		SM_TEST(r == SM_SUCCESS);
		do {
			r = select(nfds, &readfds, NULL, &readfds, &timeout);
			if (r > 0 && FD_ISSET(fd, &readfds)) {
				/* read something... */
				rd = read(fd, rdbuf, sizeof(rdbuf));
			}
		} while (r == -1 && errno == EINTR);
		if (r == 0 || r == -1) {
			close(fd);
			return 1;
		}
	}

	if (slow > 0) {
		int i;

		for (i = 0; i < sizeof(buf) - 1; i++) {
			sleep(slow / sizeof(buf));
			ret = write(fd, buf + i, 1);
			SM_TEST(1 == ret);
		}
	}
	else {
		ret = write(fd, buf, sizeof(buf));
		SM_TEST(ret == sizeof(buf));
	}
	close(fd);
	return 0;
}

int
main(int argc, char *argv[])
{
	bool clt, any;
	int family;
	int c, port, readfirst, ctmo, slow;
	sm_ret_T res;

	opterr = 0;
	clt = true;
	any = false;
	family = AF_INET;
	port = SM_DEFPORT;
	Verbose = 0;
	readfirst = 0;
	ctmo = 0;
	slow = 0;
	while ((c = getopt(argc, argv, "6c:p:r:s:uU:V")) != -1) {
		any = true;
		switch (c) {
#if MTA_NETINET6
		  case '6':
			family = AF_INET6;
			break;
#endif
		  case 'c':
			ctmo = atoi(optarg);
			if (ctmo <= 0)
				usage(argv[0]);
			break;
		  case 'p':
			port = atoi(optarg);
			if (port <= 0)
				usage(argv[0]);
			break;
		  case 'r':
			readfirst = atoi(optarg);
			break;
		  case 's':
			slow = atoi(optarg);
			if (slow <= 0)
				usage(argv[0]);
			break;
		  case 'u':
			family = AF_UNIX;
			break;
		  case 'U':
			nsocket = strdup(optarg);
			if (nsocket == NULL) {
				fprintf(stderr, "out of memory");
				return 1;
			}
			family = AF_UNIX;
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(argv[0]);
			return 1;
		}
	}
	sm_test_begin(argc, argv, "test connection to localhost by writing QUIT");
	res = net_startup();
	SM_TEST(sm_is_success(res));
	if (!sm_is_success(res))
		return sm_test_end();
	c = client(port, family, readfirst, ctmo, slow);
	net_shutdown();
	sm_test_end();
	return c;
}
